/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Stuart Parmenter <stuart@mozilla.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "nsJP2Decoder.h"

#include "imgIContainerObserver.h"

#include "nsIComponentManager.h"
#include "nsIInputStream.h"

#include "nspr.h"
#include "nsCRT.h"
#include "gfxColor.h"
#include "ImageLogging.h"
#include "nsIImage.h"
#include "nsIInterfaceRequestorUtils.h"

NS_IMPL_ISUPPORTS1(nsJP2Decoder, imgIDecoder)

#if defined(PR_LOGGING)
PRLogModuleInfo *gJP2log = PR_NewLogModule("JP2Decoder");
#else
#define gJP2log
#endif

// TODO: remove printf debugging calls

nsJP2Decoder::nsJP2Decoder()
{
	printf("nsJP2Decoder called! yay!\n");
  mBuffer = nsnull;
  mBufferSize = mBufferLen = 0;
}

nsJP2Decoder::~nsJP2Decoder()
{
  PR_FREEIF(mBuffer);
}


/** imgIDecoder methods **/

/* void init (in imgILoad aLoad); */
NS_IMETHODIMP nsJP2Decoder::Init(imgILoad *aLoad)
{
  mImageLoad = aLoad;
  mObserver = do_QueryInterface(aLoad);
  
  jas_init();

  mJASStream = jas_stream_memopen(nsnull, -1); /* let jasper handle stream memory */

  mState = JP2_START;
  
  return NS_OK;
}


/* void close (); */
NS_IMETHODIMP nsJP2Decoder::Close()
{
  PR_LOG(gJP2log, PR_LOG_DEBUG,
         ("[this=%p] nsJP2Decoder::Close\n", this));

  if(mFrame)
    mObserver->OnStopFrame(nsnull, mFrame);
  if(mImage)
    mObserver->OnStopContainer(nsnull, mImage);
  
  mObserver->OnStopDecode(nsnull, NS_OK, nsnull);

  if(mJASImage)
    jas_image_destroy(mJASImage);
    
  if(mJASStream)
    jas_stream_close(mJASStream);

  return NS_OK;
}

/* void flush (); */
NS_IMETHODIMP nsJP2Decoder::Flush()
{
  printf("nsJP2Decoder::Flush, jas stream is %p, len %d", mJASStream, jas_stream_length(mJASStream));
  int fmtid, width, height, depth, nc;
  int fw, fh;
      
  jas_stream_flush(mJASStream);
  jas_stream_rewind(mJASStream);
  
  if ((fmtid = jas_image_getfmt(mJASStream)) < 0) {
    PR_LOG(gJP2log, PR_LOG_DEBUG,("\tnsJP2Decoder::could not get image fmt\n", this));
    return NS_OK;
  }

  /* Decode the image. */
  if (!(mJASImage = jas_image_decode(mJASStream, fmtid, 0))) {
    PR_LOG(gJP2log, PR_LOG_DEBUG,("\tnsJP2Decoder::could not load image\n", this));
    return NS_OK;
  }

  nc = jas_image_numcmpts(mJASImage);
  width = jas_image_cmptwidth(mJASImage, 0);
  height = jas_image_cmptheight(mJASImage, 0);
  depth = jas_image_cmptprec(mJASImage, 0);
  
  mImage = do_CreateInstance("@mozilla.org/image/container;1");
  if (!mImage) {
    mState = JP2_ERROR;
    return NS_ERROR_OUT_OF_MEMORY;
  }
  
  mImageLoad->SetImage(mImage);
  mImage->Init(width, height, mObserver);
  mImage->GetFrameAt(0, getter_AddRefs(mFrame));
  
  mObserver->OnStartContainer(nsnull, mImage);

  
 // mObserver->OnStartDecode(nsnull);
  if(mFrame) {
    mFrame->GetWidth(&fw);
    mFrame->GetHeight(&fh);
printf("w: %d, h: %d, fw: %d, fh: %d\n", width, height, fw, fh);
    
  }
  
  if(!mFrame || fw != width || fh != height) {
    mFrame = do_CreateInstance("@mozilla.org/gfx/image/frame;2");
    if (!mFrame) {
      mState = JP2_ERROR;
      return NS_ERROR_OUT_OF_MEMORY;
    }

    gfx_format format = gfxIFormats::RGB;
#if defined(XP_WIN) || defined(XP_OS2) || defined(XP_BEOS)
    format = gfxIFormats::BGR;
#endif

    if (NS_FAILED(mFrame->Init(0, 0, width, height, format, 24))) {
      mState = JP2_ERROR;
      return NS_ERROR_OUT_OF_MEMORY;
    }

    mImage->AppendFrame(mFrame);
  }
  
  
  mObserver->OnStartFrame(nsnull, mFrame);
  
  printf("\n\nJAS:: nc %d, w/h %d/%d, depth %d, size: %ld\n", nc, width, height, depth, (long) jas_image_rawsize(mJASImage));
  
  printf("\ncolor model: %d\ncomponent types: %d %d %d\n", jas_image_clrspc(mJASImage),
         jas_image_cmpttype(mJASImage, 0),jas_image_cmpttype(mJASImage, 1),jas_image_cmpttype(mJASImage, 2));
  
  printf("\ncomponent 0:\nprec: %d\nwidth: %d ; height: %d\n", jas_image_cmptprec(mJASImage, 0),
          jas_image_cmptwidth(mJASImage, 0), jas_image_cmptheight(mJASImage, 0));
  
  mImageWidth = width;
  mImageHeight = height;
  
  WriteImage();

  return NS_OK;
}

/* unsigned long writeFrom (in nsIInputStream inStr, in unsigned long count); */
NS_IMETHODIMP nsJP2Decoder::WriteFrom(nsIInputStream *inStr, PRUint32 count, PRUint32 *_retval)
{
  LOG_SCOPE_WITH_PARAM(gJP2log, "nsJP2Decoder::WriteFrom", "count", count);
  int ret = 0;
  
  if (inStr) {
    /* Make sure we have space for at least |count| characters */
    mBuffer = (char *)PR_Realloc(mBuffer, count);
    PR_LOG(gJP2log, PR_LOG_DEBUG,("\tnsJP2Decoder::realloced\n", this));
    mBufferSize = count;

    nsresult rv = inStr->Read((char*)mBuffer, count, &mBufferLen);
    *_retval = mBufferLen;

    NS_ASSERTION(NS_SUCCEEDED(rv), "nsJP2Decoder::WriteFrom -- inStr->Read failed");
  
    jas_stream_write(mJASStream, mBuffer, count);
    jas_stream_flush(mJASStream);

    /* JPEG 2000 makes it hard to do incremental rendering.
     *   For now, we'll content ourselves with waiting until
     *   Flush() gets called before we draw the image */ 
  }
  // else no input stream.. Flush() ?

  PR_LOG(gJP2log, PR_LOG_DEBUG,
         ("[this=%p] nsJP2Decoder::WriteFrom -- processing JPEG 2000 data\n", this));

  return NS_OK;
}



PRBool
nsJP2Decoder::WriteImage()
{
  PRBool rv = PR_TRUE;
  int scanline = 0;
  // we're thebes. we can write stuff directly to the data
  PRUint8 *imageData;
  PRUint32 imageDataLength;
  if(!mFrame) {
    printf("----eep no mframe!-----\n");
    return NS_OK;
  }
  mFrame->GetImageData(&imageData, &imageDataLength);
  nsCOMPtr<nsIImage> img(do_GetInterface(mFrame));

  jas_matrix_t *mr = jas_matrix_create(mImageWidth, mImageHeight),
               *mg = jas_matrix_create(mImageWidth, mImageHeight),
               *mb = jas_matrix_create(mImageWidth, mImageHeight);
  jas_image_readcmpt(mJASImage, JAS_IMAGE_CT_RGB_R, 0, 0, mImageWidth, mImageHeight, mr);
  jas_image_readcmpt(mJASImage, JAS_IMAGE_CT_RGB_G, 0, 0, mImageWidth, mImageHeight, mg);
  jas_image_readcmpt(mJASImage, JAS_IMAGE_CT_RGB_B, 0, 0, mImageWidth, mImageHeight, mb);

  printf("extracted components...\nw: %d, h: %d\n\n",jas_matrix_numcols(mr), jas_matrix_numrows(mr));

  while (++scanline <= mImageHeight) {
      // offset is in Cairo pixels (PRUint32)
      PRUint32 offset = (scanline - 1) * mImageWidth;
      PRUint32 *ptrOutputBuf = ((PRUint32*)imageData) + offset;
    
      for (PRUint32 i=0; i < mImageWidth; ++i) {
        PRUint8 r = jas_matrix_get(mr, scanline-1, i);
        PRUint8 g = jas_matrix_get(mg, scanline-1, i);
        PRUint8 b = jas_matrix_get(mb, scanline-1, i);
        
        *ptrOutputBuf++ = GFX_PACKED_PIXEL(0xFF, r, g, b);
      }
  }

  nsIntRect r(0, 0, mImageWidth, mImageHeight);
  img->ImageUpdated(nsnull, nsImageUpdateFlags_kBitsChanged, &r);
  mObserver->OnDataAvailable(nsnull, mFrame, &r);
  
  jas_matrix_destroy(mr);
  jas_matrix_destroy(mg);
  jas_matrix_destroy(mb);
  
  return rv;
}
