/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Stuart Parmenter <pavlov@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef nsJP2Decoder_h__
#define nsJP2Decoder_h__

#include "imgIDecoder.h"

#include "nsCOMPtr.h"

#include "imgIContainer.h"
#include "gfxIImageFrame.h"
#include "imgIDecoderObserver.h"
#include "imgILoad.h"
#include "nsIInputStream.h"
#include "nsIPipe.h"


extern "C" {
#include "jasper.h"
}


#define NS_JP2DECODER_CID \
{ /* 55554fd3-6c30-4650-8735-94f271be091f */   \
     0x55554fd3,                                     \
     0x6c30,                                         \
     0x4650,                                         \
    {0x87, 0x35, 0x94, 0xf2, 0x71, 0xbe, 0x09, 0x1f} \
}

typedef enum {
  JP2_START,
  JP2_DOWNLOADING,
  JP2_DONE,

  JP2_ERROR
} jp2state;

class nsJP2Decoder : public imgIDecoder
{
public:
  NS_DECL_ISUPPORTS
  NS_DECL_IMGIDECODER

  nsJP2Decoder();
  virtual ~nsJP2Decoder();

protected:
  PRBool WriteImage();

public:
  nsCOMPtr<imgIContainer> mImage;
  nsCOMPtr<imgILoad> mImageLoad;
  nsCOMPtr<gfxIImageFrame> mFrame;

  nsCOMPtr<imgIDecoderObserver> mObserver;
  
  PRUint32 mImageWidth;
  PRUint32 mImageHeight;
  
  char *mBuffer;
  PRUint32 mBufferLen;  // amount of data currently in mBuffer
  PRUint32 mBufferSize; // size in bytes what mBuffer was created with

  jas_stream_t *mJASStream;
  jas_image_t *mJASImage;

  jp2state mState;
};

#endif // nsJP2Decoder_h__
